#include <lib_aci.h>
#include <SPI.h>
#include <EEPROM.h>
#include <CapacitiveSensor.h>
#include <notif.h>
#include "U8glib.h"

#define H1POS (0)  // Position to draw the clock (24h), H1 first number in hour, M1 first number in Minute and so on. 
#define H2POS (32)  
#define M1POS (71)
#define M2POS (102)

#define FREQ  (8000000) // CPU frequncy
#define INT_FREQ  (1)   // Frequency for the timer interupt 1Hz
//#define DEBUG         // Uncomment to initializing Serial communication

unsigned int second;    // Varibles to hold the time
unsigned int minute;
unsigned int hour;

/*
   State keep track of what is seen on the screen, and what to change to when a button is touched.
 state=0  - Screen is sleeping, bluetooth still runs, wakes (go to prev state) upp when middle button is touched, or go to message (2) for new notification.
 state=1  - Screen is on showing the time, go to sleep (0) after x seconds, go to meassage (2) if new notification or x is toched.
 state=2  - Screen showing new message, go to sleep (0) after x seconds, go to time (1) if x is touched.
 */
unsigned char state;             // Variables to handle which state the program is in
unsigned char prev_state;        // Previous state to wake up in

unsigned char sleep_timer;       // A time which the screen should go in to sleep mode
unsigned char new_touch_flag=0;  // Flag is set to 1 when a button has been touched
unsigned char new_notis_flag=0;  // Flag is set to one when a new notis is avaible 
unsigned char new_time_flag = 1; // Flag is set to 1 when a new time is avaible

U8GLIB_SSD1306_128X32 u8g(U8G_I2C_OPT_NONE);  // I2C, for the screen

/*
* Currently is only the middle and right touch button in used for this watch.
 * Uncomment the other if to be used.
 */
CapacitiveSensor cap_right = CapacitiveSensor(4, 5); // Set up Right touch button
//Capacitivensor cap_left = CapacitiveSensor(4, 8);    // Set up Left touch button
CapacitiveSensor cap_mid = CapacitiveSensor(4, 7);     // Set up Middle touch button

long val_right;     // Variables to store capacivitive value for right touch button
//long val_left;    // Variables to store capacivitive value for left touch button
long val_mid;       // Variables to store capacivitive value for middle touch button


Notif notif(10, 2);      // Start Notif and select which pins used
ancs_notification_t* current_notif = NULL;  // An object to hold the notification

boolean connected = false;  // Boolean to se if any device is connected


/*
* Functions for the bluetooth module to work properly. 
 */
void ancs_connected() {
  connected = true;
}
void ancs_disconnected() {
  connected = false;
}

void ancs_reset() {
  connected = false;
}

/*
* Function called when a new notification appears. Store the notis to current_notis.
 * Update time from iPhone every time a new notis appears. 
 * Set new_nots_flag to 1
 */
void ancs_notifications(ancs_notification_t* notif) {
  current_notif = notif;
  hour=current_notif->time.h;
  minute=current_notif->time.m;
  second=current_notif->time.s;
  new_notis_flag=1;
}

//If things get really crazy, uncomment this line. It wipes the saved EEPROM information for the Nordic chip. Good to do this if the services.h file gets updated.
//After it is wiped, comment and reupload.
/*
  void eepromWrite(int address, uint8_t value)
 {
 eeprom_write_byte((unsigned char *) address, value);
 }
 */



/*
* Start timer 1 with a frequency of INT_FREQ
 */
void init_timer() {
  cli();//stop interrupts
  //set timer1 interrupt at 1Hz
  TCCR1A = 0;// set entire TCCR1A register to 0
  TCCR1B = 0;// same for TCCR1B
  TCNT1  = 0;//initialize counter value to 0
  // set compare match register for 1hz increments
  OCR1A = FREQ / 1024 / INT_FREQ; // (must be <65536)
  // turn on CTC mode
  TCCR1B |= (1 << WGM12);
  // Set CS12 and CS10 bits for 1024 prescaler
  TCCR1B |= (1 << CS12) | (1 << CS10);
  // enable timer compare interrupt
  TIMSK1 |= (1 << OCIE1A);

  sei();//allow interrupts
}

/*
* Update sleep timer, update for which second the screen should go to sleep.
 * Screen goes to sleep after 10 seconds from any event.
 */
void update_sleep(){
  if (second < 50) {
    sleep_timer = second + 10;
  } 
  else {
    sleep_timer = 10 - (60 - second);
  }
}

/*
* Update the screen regarding which state present.
 * state=0 -> Go to sleep
 * state=1 -> Draw the clock, update sleep timer
 * state=2 -> Show the message, update sleep timer
 */
void update_screen(){
  if(state==0){
    u8g.sleepOn();
    sleep_timer=60;  // Second will neveer reach 60, will stay in state zero undtil new event apperas.
  }
  else if (state==1){
    u8g.firstPage();
    do {
      draw_clock(hour, minute);
    } 
    while ( u8g.nextPage() );
    delay(50);
    new_time_flag=0;
    update_sleep();
  }
  else if (state==2){
    u8g.firstPage();
    do {
      message_handler();
    } 
    while ( u8g.nextPage() );
    delay(50); 
    new_notis_flag=0;
    update_sleep();
  } 
}

/*
* Function to reset the arduino. Touch middle and right at the same time to reset.
*/
void(* resetFunc) (void) = 0; //declare reset function @ address 0

void setup() {

  //If things get really crazy, uncomment this line. It wipes the saved EEPROM information for the Nordic chip. Good to do this if the services.h file gets updated.
  //After it is wiped, comment and reupload.
  // eepromWrite(0, 0xFF);

#ifdef DEBUG
  Serial.begin(115200);
#endif

  u8g.setFont(u8g_font_profont10);  // Sets the font for the message to be written in. 

  hour = 0;
  minute = 0;
  second = 0;
  init_timer();

  notif.setup();
  notif.set_notification_callback_handle(ancs_notifications);
  notif.set_connect_callback_handle(ancs_connected);
  notif.set_disconnect_callback_handle(ancs_disconnected);
  notif.set_reset_callback_handle(ancs_reset);

  sleep_timer=60;
  state = 1;
  prev_state=state;
  update_screen();
}


void loop() {
  notif.ReadNotifications();  // This handles the bluetooth connection and receiving notifications
  val_mid = cap_mid.capacitiveSensor(3);  // the value 3 might be change, test to find a better value.

  if (val_mid > 10) {  // The value 10 might be changed, test to find a better value. 
    new_touch_flag=1;  // set new_touch_flag=1 if the middle button is touched. 
  }

/*
* Handle and change state when the middle button is pressed.
*/
  
  if(new_touch_flag==1){
    switch (state) {
    case 0:
      state = prev_state;
      u8g.sleepOff();
      update_screen();

      break;
    case 1:
      prev_state = state;
      state = 2;
      update_screen();
      break;
    case 2:
      prev_state = state;
      state = 1;
      update_screen();
      break;
    }
    new_touch_flag=0;
    
    delay(200);
    val_right=cap_right.capacitiveSensor(3);
    if(val_right>10){
      resetFunc();  //Call reset, if both the middle and the right button are touched at the same time. 
    }
  }

/*
* Handle the sleep timer, set state to 0 and put the screen in sleep mode. 
*/
  if(second==sleep_timer){
    prev_state=state;
    state=0;
    update_screen();
  } 

/*
* Handles the new recieved notification,set state to 2, wake up screen and update screen with new message. 
*/
  if((state==0 || state==1 )&& new_notis_flag==1){
    state=2; 
    u8g.sleepOff();  
    update_screen();
  }
/*
* Update the time on the screen if state=1.
*/
  if(state==1 && new_time_flag==1){
    update_screen();
  }

}


/*
   Timer 1, this function execute with a frequency of 1 Hz. This increases the variable second.
 If second >59 then minute is changed and so on. This function handles the clock.
 A flag "new_time_flag" is set to 1 to indicate that a the time has updated.
 */

ISR(TIMER1_COMPA_vect) {
  if (second < 59) {
    second++;
  } 
  else if (minute < 59) {
    second = 0;
    minute++;
    new_time_flag = 1;
  } 
  else if (hour < 23) {
    second = 0;
    minute = 0;
    hour++;
    new_time_flag = 1;
  } 
  else {
    second = 0;
    minute = 0;
    hour = 0;
    new_time_flag = 1;
  }
}






